import sys
from datetime import datetime

import click

from .supervisor import AsyaSupervisor
from .logic import gather_acquaintances
from .exceptions import AsyaException


_user_involvement = ('author', 'involves', 'mentions',
                     'assigned', 'commenter')

API_SEARCH_LIMIT = 1000


def create_search_specs(username, sort, order, text, involvement, query_opts):
    """Create GitHub issues search specification (params dict)"""
    query_opts[involvement] = username
    query = '+'.join(['{}:{}'.format(k, v)
                      for k, v in query_opts.items()
                      if v is not None])
    if text is not None:
        query = '+'.join([text, query])

    search_specs = {'q': query}
    if sort is not None:
        search_specs['sort'] = sort
    if order is not None:
        search_specs['order'] = order

    return search_specs


def setup_progressbar(supervisor):
    """Setup progressbar for given supervisor"""
    supervisor.data['skipped'] = 0

    def add_skipped(headers):
        supervisor.data['skipped'] += 1

    def report_skipped():
        if supervisor.data['skipped'] > 0:
            click.echo('{} error(s) 404 skipped'.format(
                supervisor.data['skipped']
            ))

    supervisor.data['waiting'] = 0
    supervisor.data['bar_message'] = None

    def show_message(item):
        if supervisor.data['bar_message'] is None:
            return ''
        return '({})'.format(supervisor.data['bar_message'])

    supervisor.data['bar'] = click.progressbar(
        length=0, label='Processing issues', show_pos=True,
        fill_char='\u2588', empty_char='\u2591',
        item_show_func=show_message,
        bar_template='%(label)s:  %(bar)s %(info)s'
    )

    def init_bar(first_result, page):
        if page == 1:
            nresults = min(API_SEARCH_LIMIT, first_result['total_count'])
            if first_result['total_count'] > API_SEARCH_LIMIT:
                click.echo('{} issue(s) will be processed (due to API limits,'
                           ' although there are {} issues)'.format(
                    API_SEARCH_LIMIT, first_result['total_count']
                ))
            supervisor.data['bar'].length = nresults
            supervisor.data['bar'].entered = True
            supervisor.data['bar'].render_progress()

    def increase_bar(issue):
        supervisor.data['bar'].update(1)

    def finish_bar():
        supervisor.data['bar'].finish()
        supervisor.data['bar'].render_finish()

    def waiting_phase_change(active, headers):
        to_time = int(headers['X-RateLimit-Reset'])
        if active and to_time > supervisor.data['waiting']:
            supervisor.data['waiting'] = to_time
            supervisor.data['bar_message'] = 'waiting [{}]'.format(
                datetime.fromtimestamp(to_time).strftime('%Y-%m-%d %H:%M:%S')
            )
            supervisor.data['bar'].render_progress()
        if not active and to_time == supervisor.data['waiting']:
            supervisor.data['waiting'] = 0
            supervisor.data['bar_message'] = None
            supervisor.data['bar'].render_progress()

    supervisor.callbacks['issues_search_page'].append(init_bar)
    supervisor.callbacks['issue'].append(increase_bar)
    supervisor.callbacks['skip'].append(add_skipped)
    supervisor.callbacks['wait'].append(waiting_phase_change)
    supervisor.callbacks['finish_successful'].append(finish_bar)
    supervisor.callbacks['finish_errored'].append(finish_bar)
    supervisor.callbacks['finish_successful'].append(report_skipped)
    supervisor.callbacks['finish_errored'].append(report_skipped)


def setup_info_msgs(supervisor):
    """Setup text info printing for given supervisor"""
    supervisor.data['waiting'] = 0
    supervisor.data['skipped'] = 0

    def add_skipped(headers):
        supervisor.data['skipped'] += 1

    def report_skipped():
        if supervisor.data['skipped'] > 0:
            click.echo('{} error(s) 404 skipped'.format(
                supervisor.data['skipped']
            ))

    def print_count(first_result, page):
        if page == 1:
            if first_result['total_count'] > API_SEARCH_LIMIT:
                click.echo('{} issue(s) will be processed (due to API limits,'
                           ' although there are {} issues)'.format(
                    API_SEARCH_LIMIT, first_result['total_count']
                ))
            else:
                click.echo('{} issue(s) to be processed...'.format(
                    first_result['total_count']
                ))

    def waiting_phase_change(active, headers):
        to_time = int(headers['X-RateLimit-Reset'])
        if active and to_time > supervisor.data['waiting']:
            supervisor.data['waiting'] = to_time
            click.echo('Waiting for API rate limit reset until {}'.format(
                datetime.fromtimestamp(to_time).strftime('%Y-%m-%d %H:%M:%S')
            ))
        if not active and to_time == supervisor.data['waiting']:
            supervisor.data['waiting'] = 0
            click.echo('Resuming working after wait')

    supervisor.callbacks['issues_search_page'].append(print_count)
    supervisor.callbacks['wait'].append(waiting_phase_change)
    supervisor.callbacks['skip'].append(add_skipped)
    supervisor.callbacks['finish_successful'].append(report_skipped)
    supervisor.callbacks['finish_errored'].append(report_skipped)


def print_result(result):
    """Print Asya result nicely"""
    if result is None:
        click.secho('Result is None!', fg='red')
        return
    elif len(result) == 0:
        click.echo('No results to print...')
        return
    d = [(k, result[k]) for k in sorted(result, key=result.get, reverse=True)]
    uwidth = max([len(username) for username, _ in d])
    cwidth = max([len(str(count)) for _, count in d])
    for username, count in d:
        click.echo(
            '{u:{uw}} = {c:{cw}}'.format(
                u=username, uw=uwidth,
                c=count, cw=cwidth
            )
        )


def no_print(*args, **kwargs):
    """Dummy method for not actually printing anything"""
    pass


@click.command()
@click.argument('username')
@click.option('-i', '--involvement', type=click.Choice(_user_involvement),
              default='author', help='How is given user involved in issues.')
@click.option('--text', help='Text to filter issues.')
@click.option('--in', type=click.Choice(['title', 'body', 'comments']),
              help='In what should be text searched.')
@click.option('--type', type=click.Choice(['issue', 'pr']),
              help='Restriction to just issues or just PRs.')
@click.option('--state', type=click.Choice(['open', 'closed']),
              help='Restriction to just issues or just PRs.')
@click.option('--created',
              help='Date expression to filter by created date.')
@click.option('--updated',
              help='Date expression to filter by updated date.')
@click.option('--label',
              help='String to filter by label.')
@click.option('--language',
              help='String to filter by language.')
@click.option('--sort', type=click.Choice(['comments', 'created', 'updated']),
              help='Sort search results (important for >1000 results).')
@click.option('--order', type=click.Choice(['asc', 'desc']),
              help='Sort order of results (important for >1000 results).')
@click.option('-t', '--token', default=None,
              envvar='GITHUB_TOKEN', help='Personal GitHub token.')
@click.option('-w', '--wait-rate-limit', is_flag=True,
              help='Wait for rate limit reset if needed.')
@click.option('-s', '--skip-404', is_flag=True,
              help='Skip not-found GitHub resources (such as disabled repos).')
@click.option('-b', '--progress-bar', is_flag=True,
              help='Toggle progress bar (default false).')
@click.option('--info/--no-info', default=True,
              help='Toggle info texts (default true).')
@click.option('--api-endpoint', default='https://api.github.com',
              help='How is given user involved in issues.')
@click.option('--debug', is_flag=True, default=False,
              help='Debug mode (not catching other exceptions).')
@click.version_option('0.1')
def main(username, token, wait_rate_limit, sort, order, progress_bar, info,
         skip_404, text, involvement, api_endpoint, debug, **query_opts):
    """Asya Command Line Interface (via :mod:`click`)"""
    search_specs = create_search_specs(username, sort, order, text,
                                       involvement, query_opts)
    supervisor = AsyaSupervisor(api_endpoint, token, wait_rate_limit, skip_404)

    print_info = no_print

    if progress_bar:
        setup_progressbar(supervisor)
    elif info:
        setup_info_msgs(supervisor)
    if info:
        print_info = click.secho

    if debug:
        result = gather_acquaintances(search_specs, supervisor)
        supervisor.report_finish_successful()
        print_info('Asya gathered acquaintances successfully:',
                   fg='green', bold=True, err=True)
        print_result(result)
    else:
        try:
            result = gather_acquaintances(search_specs, supervisor)
            supervisor.report_finish_successful()
            print_info('Asya gathered acquaintances successfully:',
                       fg='green', bold=True, err=True)
            print_result(result)
        except AsyaException as err:
            supervisor.report_finish_errored()
            print_info('Asya ended with communication error:',
                       fg='red', bold=True, err=True)
            click.echo(err.message, err=True)
            sys.exit(7)
        except Exception as err:
            supervisor.report_finish_errored()
            print_info('Asya ended with fatal error:',
                       fg='red', bold=True, err=True)
            click.echo(err.__class__.__name__, err=True)
            sys.exit(10)



from collections import defaultdict


class AsyaSupervisor:
    """
    Supervisor for running the Asya gathering async procedure.
    It contains some procedure-wide setting and calls.

    The ``gather_acquaintances`` function must call supervisor's
    ``report_`` methods at the right places (described in docstrings):

    .. code ::

       ...
       # receiving and processing the issue
       supervisor.report_issue(issue)
       # additional work with the issue
       ...

    Your implementation may add custom callbacks:

    .. code ::

       def my_procedure(issue):
           ...

       supervisor.callbacks['issue'].append(my_procedure)
       # my_procedure will be called when report_issue is
       # called on the supervisor object

    You may also use ``supervisor.obj`` for your data as you
    need. Do not touch other parameters nor edit this class!

    :ivar api_endpoint: API endpoint to be used for communication
    :ivar token: API token to be used (you can use ``has_token``)
    :ivar wait_rate_limit:
        True if the app should wait until the rate limit resets after it is
        exceeded, False if exceeding the limit should cause
        :exc:`~asya.exceptions.AsyaException`
    :ivar skip_404:
        True the app should skip 404 errors, False if they should
        raise :exc:`~asya.exceptions.AsyaException`
    :ivar per_page: size of page for API requests
    """

    def __init__(self, api_endpoint, token, wait_rate_limit, skip_404,
                 per_page=100):
        self.api_endpoint = api_endpoint
        self.token = token
        self.wait_rate_limit = wait_rate_limit
        self.skip_404 = skip_404
        self.per_page = per_page
        self.data = {}
        self.callbacks = defaultdict(list)
        self.obj = None  #: User obj (can be anything)

    @property
    def has_token(self):
        """True if the API token is set"""
        return self.token is not None

    def report_issues_search_page(self, page, number):
        """
        Method to be called before processing of a single result page

        :param page: page of issues search (data from API)
        :type page: dict
        :param number: number of the page
        :type number: int
        """
        self._do_callback('issues_search_page', page, number)

    def report_issue(self, issue):
        """
        Method to be called after processing of a single issue

        :param issue: GitHub issue (data from API)
        :type issue: dict
        """
        self._do_callback('issue', issue)

    def report_comment(self, comment):
        """
        Method to be called after processing of a single comment

        :param comment: GitHub comment (data from API)
        :type comment: dict
        """
        self._do_callback('comment', comment)

    def report_wait(self, active, headers):
        """
        Method to be called whenever changing waiting state
        (see ``wait_rate_limit``)

        :param active: True if wait is starting, False if ending
        :type active: bool
        :param headers: headers of the API response that caused waiting
        :type headers: dict
        """
        self._do_callback('wait', active, headers)

    def report_skip(self, headers):
        """
        Method to be called whenever some result is skipped
        (see ``skip_404``)

        :param headers: headers of the API response that caused skipping
        :type headers: dict
        """
        if not self.skip_404:
            raise PermissionError('Not allowed to skip anything!')
        self._do_callback('skip', headers)

    def _do_callback(self, callback_name, *args, **kwargs):
        for call in self.callbacks[callback_name]:
            call(*args, **kwargs)

    def __getattr__(self, item):
        if item.startswith('report_'):
            callback_name = item[7:]

            def callback(*args, **kwargs):
                return self._do_callback(callback_name, *args, **kwargs)
            return callback

        raise AttributeError(item)




class AsyaException(Exception):
    """
    Exception of Asya caused during communication with
    the GitHub API

    :ivar data: API response data causing the exception
    :vartype data: dict
    :ivar headers: headers of the API response that caused the exception
    :vartype headers: dict
    """

    def __init__(self, data, headers):
        self.data = data
        self.headers = headers

    @property
    def message(self):
        return '{}: {}'.format(
            self.headers['Status'],
            self.data.get('message', '')
        )



from setuptools import setup, find_packages

with open('README.rst') as f:
    long_description = ''.join(f.readlines())

setup(
    name='asya',
    version='0.1',
    keywords='github issues social search asyncio comments issues',
    description='Asynchronously your acquaintances from GitHub issues',
    long_description=long_description,
    author='Marek Suchánek',
    author_email='suchama4@fit.cvut.cz',
    license='MIT',
    url='https://github.com/MarekSuchanek/asya',
    packages=find_packages(),
    entry_points={
        'console_scripts': [
            'asya = asya:main',
        ]
    },
    install_requires=[
        'aiohttp',
        'requests',
        'click'
    ],
    python_requires='>=3.4',
    classifiers=[
        'Development Status :: 1 - Planning',
        'Framework :: AsyncIO',
        'Intended Audience :: Developers',
        'Intended Audience :: Education',
        'Intended Audience :: Information Technology',
        'Intended Audience :: Science/Research',
        'License :: OSI Approved :: MIT License',
        'Natural Language :: English',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Topic :: Communications',
        'Topic :: Sociology',
    ],
)



import asyncio
import async_timeout
import time

import aiohttp
import requests
from urllib.parse import urlparse, parse_qs

from .exceptions import AsyaException


class GitHubHeaders:

    def __init__(self, headers):
        self.headers = headers

    def __getitem__(self, key):
        return self.headers.get(key, '')

    @property
    def status_code(self):
        return int(self.headers['Status'][0:3])

    @property
    def ratelimit_exhausted(self):
        return self.status_code == 403 and self.ratelimit_remaining == 0

    @property
    def ratelimit_remaining(self):
        return int(self.headers.get('X-RateLimit-Remaining', -1))

    @property
    def ratelimit_reset(self):
        return int(self.headers.get('X-RateLimit-Reset', -1))

    @property
    def ratelimit_wait(self):
        return max(0, self.ratelimit_reset - int(time.time()))


def get_last_page(headers):
    links = requests.utils.parse_header_links(headers['Link'])
    links2 = {link['rel']: link['url'] for link in links}
    params = parse_qs(urlparse(links2['last']).query)
    return int(params['page'][0])


async def fetch_data_header(session, url, params=None):
    with async_timeout.timeout(10):
        async with session.get(url, headers=session.headers,
                               params=params) as response:
            return await response.json(), response.headers


async def fetch_and_process(supervisor, session, url, processor, params=None,
                            expected_code=200):
    data, headers = await fetch_data_header(session, url, params)

    gheaders = GitHubHeaders(headers)
    if gheaders.status_code != expected_code:
        if gheaders.status_code == 404 and supervisor.skip_404:
            supervisor.report_skip(headers)
            return None

        if gheaders.ratelimit_exhausted and supervisor.wait_rate_limit:
            supervisor.report_wait(True, headers)
            await asyncio.sleep(gheaders.ratelimit_wait)
            supervisor.report_wait(False, headers)
            return await fetch_and_process(supervisor, session, url,
                                           processor, params, expected_code)

        raise AsyaException(data, headers)
    return await processor(data, headers)


async def process_comment(supervisor, comment):
    supervisor.report_comment(comment)


async def process_pages(session, supervisor, url, params, process_item):
    async def process(data, headers):
        async def process_page(page):
            xparams = {'page': page}
            xparams.update(params)

            async def process_data(data, _):
                futures = [
                    asyncio.ensure_future(process_item(item)) for item in data
                ]
                await asyncio.gather(*futures)

            await fetch_and_process(supervisor, session, url,
                                    process_data, xparams)

        futures = [asyncio.ensure_future(process_item(item)) for item in data]

        if 'Link' in headers:
            last_page = get_last_page(headers)
            for act_page in range(2, last_page + 1):
                futures.append(asyncio.ensure_future(process_page(act_page)))

        await asyncio.gather(*futures)

    await fetch_and_process(supervisor, session, url, process, params)


async def process_issue(session, supervisor, issue):
    async def process(comment):
        await process_comment(supervisor, comment)

    await process_pages(
        session,
        supervisor,
        issue['comments_url'],
        {'per_page': supervisor.per_page},
        process
    )
    supervisor.report_issue(issue)


async def gather_acquaintances_from_issues(search_specs, supervisor):
    url = supervisor.api_endpoint + '/search/issues'
    async with aiohttp.ClientSession() as session:
        session.headers = {'User-Agent': 'Python/ASYA'}
        if supervisor.has_token:
            session.headers['Authorization'] = 'token ' + supervisor.token

        async def process_item(issue):
            await process_issue(session, supervisor, issue)

        async def process_page(data, _, page):
            supervisor.report_issues_search_page(data, page)
            futures = [asyncio.ensure_future(process_item(item))
                       for item in data['items']]
            await asyncio.gather(*futures)

        async def process_query_page(page):
            params = {'page': page}
            params.update(search_specs)

            async def process_data(data, headers):
                await process_page(data, headers, page)

            await fetch_and_process(supervisor, session, url,
                                    process_data, params)

        async def process(data, headers):
            futures = [asyncio.ensure_future(process_page(data, headers, 1))]

            if 'Link' in headers:
                last_page = get_last_page(headers)
                for page in range(2, last_page + 1):
                    futures.append(asyncio.ensure_future(
                        process_query_page(page))
                    )

            await asyncio.gather(*futures)

        await fetch_and_process(supervisor, session, url,
                                process, search_specs)


def gather_acquaintances(search_specs, supervisor):
    """Gather acquaintances from GitHub issues and comments with
    given search_specs with counts of comments in form of dict.
    It uses :mod:`asyncio` and :class:`aiohttp.ClientSession`.

    >>> gather_acquaintances({'q': 'author:MarekSuchanek'}, supervisor)
    {'MarekSuchanek': 7, 'hroncok': 15, 'encukou': 10}

    For more information about the ``search_specs`` content visit the
    `GitHub Search API docs <https://developer.github.com/v3/search/#search-issues>`_.

    :param search_specs: dictionary with search specification (params for the search)
    :type search_specs: dict
    :param supervisor: supervisor object used for this gathering
    :type supervisor: asya.supervisor.AsyaSupervisor

    :return: dictionary with usernames as keys and number of comments as values
    :rtype: dict
    """
    # return {}
    from collections import defaultdict

    supervisor.obj = {'counts': defaultdict(int)}

    search_specs['per_page'] = supervisor.per_page

    def add_comment(comment):
        supervisor.obj['counts'][comment['user']['login']] += 1

    supervisor.callbacks['comment'].append(add_comment)

    loop = asyncio.get_event_loop()
    loop.run_until_complete(
        gather_acquaintances_from_issues(search_specs, supervisor)
    )
    loop.close()

    return supervisor.obj['counts']
